#ifndef _XewBasic_h
#define _XewBasic_h
/*
** Copyright 1992 by Markku Savela and
**	Technical Research Centre of Finland
**
** The author maybe contacted via:
**
**    Mail:	Markku Savela
**		Technical Research Centre of Finland
**		Telecommunications Laboratory
**		Otakaari 7 B, P.O. Box 34
**		SF-02151 ESPOO
**		Finland
**
**    Phone:	+358 0 456 6509
**    Email:	savela@tel.vtt.fi
*/

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Converters.h>

/* Parameters:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 border		     BorderColor	Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	1
 contentFile         ContentFile        String          NULL
 contentFormat       ContentFormat      XeContentFormat XeContentFormat_UNKNOWN
 contentLoaded       ContentLoaded      Boolean         False
 contentLength       ContentLength      long            0
 contentOffset       ContentOffset      long            0
 contentStream       ContentStream      FILE *          NULL
 contentString       ContentString      Pointer         NULL
 destroyCallback     Callback		Pointer		NULL
 cursor              Cursor             Cursor          None
 foreground	     Foreground		Pixel		XtDefaultForeground
 height	             Height		Dimension	0
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 notifyCallback      NotifyCallback     Pointer         NULL
 resize              Resize             Boolean         True
 scaling             Scaling            XeScaling       XeScaling_WIDTH
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
*/

typedef enum
    {
	XeContentFormat_UNKNOWN,/* Unknown format */
	XeContentFormat_ISO2022,/* ISO 2022 octet stream */
	XeContentFormat_TIFF,	/* TIFF 6.0 Base line */
	XeContentFormat_JPEG,	/* JPEG/JFIF stream */
	XeContentFormat_GIF,	/* GIF89a stream */
	XeContentFormat_CGM,	/* CGM octet stream */
	XeContentFormat_H261,	/* H.261 octet stream */
	XeContentFormat_AUDIO,	/* 8kHz u-law stream */
	XeContentFormat_ODIF	/* ODA/ODIF stream */
    } XeContentFormat;

typedef enum
    {
	XeScaling_NONE,		/* No scaling, visualize in real size */
	XeScaling_WIDTH,	/* Width controlled, preserve aspect ratio */
	XeScaling_HEIGHT,	/* Height controlled, preserve aspect ratio */
	XeScaling_MAXASPECT,	/* Max size withing window, preserve aspect */
	XeScaling_FITAREA	/* Fit exactly into window, change aspect */
    } XeScaling;

/*
** Known callback reasons in Xe widgets
*/
typedef enum
    {
	XeCR_UNKNOWN,		/* Unspecified callback reason */
	XeCR_NOTIFY,		/* Universal default reason for notify() */
	XeCR_NOTIFY_AREA	/* Area of interest report */
    } XeCR;

/*
** XeCallData
**	is the general structure that is passed to all Xew callbacks in
**	Xew. Individual widget classes may have further extra fields and
**	have an alternate definition for the structure and different
**	callbacks, but the beginning always matches this structure.
**
**	NOTE:	The callback reason codes will be globally unique across
**		all Xe widgets (the reason code alone can be used to decide
**		about the actual structure that applies to call_data,
**		regardless of the source widget of the callback).
*/
typedef struct XeCallbackData
    {
	int reason;		/* Reason of the callback */
    } XeCallbackData;

/*
** XeNotifyCallData
**	is the universal default data for the notify callback.
*/
typedef struct XeNotifyCallbackData
    {
	int reason;		/* == XeCR_NOTIFY == */
	XEvent *event;		/* event that triggered the callback */
	String *params;		/* Point to args of 'notity(args)' */
	Cardinal num_params;	/* Number of arguments in 'args' */
    } XeNotifyCallbackData;

/*
** XeNotifyAreaCallbackData
**	is the data supplied for the XeCR_NOTIFY_AREA callback. This
**	callback is made via general XtNnotifyCallback. The widget uses
**	this callback when it detects that the area of interest shifts
**	outside the current parent window pounds (like moving the cursor
**	out of displayed area in Text Widget). This information can be
**	used to control vieports, portholes and similar constructs.
*/
typedef struct XeNotifyAreaCallbackData
    {
	int reason;		/* == XeCR_NOTIFY_AREA == */
	XRectangle area;	/* Area of interest */
	XRectangle visible;	/* Visiblea are of the window */
    } XeNotifyAreaCallbackData;


/* New fields */
#define XtNcursor "cursor"
#define XtNcontentFile "contentFile"
#define XtCContentFile "ContentFile"
#define XtNcontentFormat "contentFormat"
#define XtCContentFormat "ContentFormat"
#define XtRXeContentFormat "XeContentFormat"
#define XtNcontentLoaded "contentLoaded"
#define XtCContentLoaded "ContentLoaded"
#define XtNcontentStream "contentStream"
#define XtCContentStream "ContentStream"
#define XtNcontentString "contentString"
#define XtCContentString "contentString"
#define XtNcontentLength "contentLength"
#define XtCContentLength "ContentLength"
#define XtNcontentOffset "contentOffset"
#define XtCContentOffset "ContentOffset"
#define XtNnotifyCallback "notifyCallback"
#define XtCNotifyCallback "NotifyCallback"
#define XtNscaling "scaling"
#define XtCScaling "scaling"
#define XtRXeScaling "XeScaling"
#define XtRHorizontalDimension "HorizontalDimension"
#define XtRVerticalDimension "VerticalDimension"

/* Class record constants */

extern WidgetClass xeBasicWidgetClass;

typedef struct _XeBasicClassRec *XeBasicWidgetClass;
typedef struct _XeBasicRec      *XeBasicWidget;


#endif
