/*
** Copyright 1992, 1993 by Markku Savela and
**	Technical Research Centre of Finland
**
** The author may be contacted via:
**
**    Mail:	Markku Savela
**		Technical Research Centre of Finland
**		Telecommunications Laboratory
**		Otakaari 7 B, P.O. Box 34
**		SF-02151 ESPOO
**		Finland
**
**    Phone:	+358 0 456 6509
**    Email:	savela@tel.vtt.fi
*/
#if SYSV_INCLUDES
#	include <memory.h>
#	include <malloc.h>
#endif
#if ANSI_INCLUDES
#	include <stddef.h>
#	include <stdlib.h>
#endif
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/CharSet.h>
#include <X11/Xew/XewInit.h>
#include <X11/Xew/RasterP.h>
#include "RasterImport.h"
#include "RasterLayout.h"
#include "RasterColor.h"

#define offset(field) XtOffsetOf(XeRasterRec, raster.field)

static XtResource resources[] =
    {
	{XtNclipX, XtCClipX, XtRLong, sizeof(long),
		 offset(clip_x), XtRImmediate, (long)0},
	{XtNclipY, XtCClipY, XtRLong, sizeof(long),
		 offset(clip_y), XtRImmediate, (long)0},
	{XtNclipWidth, XtCClipWidth, XtRLong, sizeof(long),
		 offset(clip_width), XtRImmediate, (long)0},
	{XtNclipHeight, XtCClipHeight, XtRLong, sizeof(long),
		 offset(clip_height), XtRImmediate, (long)0},
	{XtNcolormapUse, XtCColormapUse, XtRXeColormapUse,
		 sizeof(XeColormapUse), offset(colormap_use),
		 XtRImmediate, (XtPointer)XeColormapUse_SHARED},
	{XtNcolormapInstall, XtCColormapInstall, XtRBoolean,
		 sizeof(Boolean),offset(colormap_install),XtRImmediate,False},
	{XtNcolorQuantize, XtCColorQuantize, XtRXeColorQuantize,
		 sizeof(XeColorQuantize), offset(color_quantize),
		 XtRImmediate, (XtPointer)XeColorQuantize_FAST},
	{XtNoffsetX, XtCOffsetX, XtRDimension, sizeof(Dimension),
		 offset(offset_x), XtRImmediate, (long)0},
	{XtNoffsetY, XtCOffsetY, XtRDimension, sizeof(Dimension),
		 offset(offset_y), XtRImmediate, (long)0},
	{XtNmaxColors, XtCMaxColors, XtRInt, sizeof(int),
		 offset(max_colors), XtRImmediate, (XtPointer)65535},
	{XtNrasterHeight, XtCRasterHeight, XtRInt,
		 sizeof(unsigned int), offset(height), XtRImmediate, 0},
	{XtNrasterWidth, XtCRasterWidth, XtRInt,
		 sizeof(unsigned int), offset(width), XtRImmediate, 0},
	{XtNrasterWidth, XtCRasterWidth, XtRInt,
		 sizeof(unsigned int), offset(width), XtRImmediate, 0},
	{XtNvisual, XtCVisual, XtRVisual,
		 sizeof(Visual *), offset(visual), XtRImmediate, 0},
    };
#undef offset

static void ClassPartInitialize(), ClassInitialize();
static void Initialize(), Redisplay(), Resize(), Realize(), Destroy();
static Boolean SetValues();

#if NeedFunctionPrototypes
static void EnterWindow(Widget, XEvent *, String *, Cardinal *);
static void LeaveWindow(Widget, XEvent *, String *, Cardinal *);
static void Notify(Widget, XEvent *, String *, Cardinal *);
#else
static void EnterWindow();
static void LeaveWindow();
static void Notify();
#endif

static char defaultTranslations[] =
"\
<Btn1Down>,<Btn1Up>: notify()\n\
<EnterWindow>:	enter-window()\n\
<LeaveWindow>:	leave-window()\
";

static XtActionsRec actions[] =
    {
	    {"notify",		Notify},
	    {"enter-window",	EnterWindow},
	    {"leave-window",	LeaveWindow},
    };



XeRasterClassRec xeRasterClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &xeBasicClassRec,
    /* class_name		*/	"XeRaster",
    /* widget_size		*/	sizeof(XeRasterRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	ClassPartInitialize,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	XtExposeCompressMultiple,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	defaultTranslations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* XeBasic fields */
    /* not used			*/	0
  },
  { /* XeRaster fields */
    /* not used			*/	0
  }
};

WidgetClass xeRasterWidgetClass = (WidgetClass)&xeRasterClassRec;

#define	CvtDone(type, value)				\
    do {						\
	if (toVal->addr != NULL)			\
	    {						\
		if (toVal->size < sizeof(type))		\
		    {					\
			toVal->size = sizeof(type);	\
			return False;			\
		    }					\
		*(type*)(toVal->addr) = (value);	\
	    }						\
	else						\
	    {						\
		static type static_val;			\
		static_val = (value);			\
		toVal->addr = (XPointer)&static_val;	\
	    }						\
	toVal->size = sizeof(type);			\
	return True;					\
    } while (0)

static CvtEnumInfo cvtColormapUse[] =
    {
	{"shared",	XeColormapUse_SHARED },
	{"sharedown",	XeColormapUse_SHAREDOWN },
	{"optional",	XeColormapUse_OPTIONAL },
	{"private",	XeColormapUse_PRIVATE },
	{NULL,		sizeof(XeColormapUse) },
    };

static CvtEnumInfo cvtColorQuantize[] =
    {
	{"slow",	XeColorQuantize_SLOW },
	{"fast",	XeColorQuantize_FAST },
	{"best",	XeColorQuantize_BEST },
	{"jpeg",	XeColorQuantize_JPEG },
	{NULL,		sizeof(XeColorQuantize) },
    };

static void ClassInitialize()
    {
	static XtConvertArgRec convertColormapUseArg[] =
	    {
		{
		    XtAddress,
		    (XtPointer) &cvtColormapUse[0],
		    sizeof(XtPointer),
		}
	    };

	static XtConvertArgRec convertColorQuantizeArg[] =
	    {
		{
		    XtAddress,
		    (XtPointer) &cvtColorQuantize[0],
		    sizeof(XtPointer),
		}
	    };

	XewInitializeWidgetSet();
	XtSetTypeConverter(XtRString, XtRXeColormapUse,
			   XeCvtStringToEnum,
			   convertColormapUseArg,
			   XtNumber(convertColormapUseArg),
			   XtCacheNone,
			   (XtDestructor)NULL);
	XtSetTypeConverter(XtRString, XtRXeColorQuantize,
			   XeCvtStringToEnum,
			   convertColorQuantizeArg,
			   XtNumber(convertColorQuantizeArg),
			   XtCacheNone,
			   (XtDestructor)NULL);
    }

static void ClassPartInitialize(class)
WidgetClass class;
    {
    }

/*
** SetupContent
**	Prepare for viewing raster content
*/
static void SetupContent(w)
XeRasterWidget w;
    {
	Dimension width, height;

	if (w->basic.content_loaded == False)
	    {
		XeRasterImport(w);
		w->basic.content_loaded = True;
		w->raster.done = False;
	    }
	XeBasicScaling((XeBasicWidget)w,
		       w->raster.width, w->raster.height,
		       &width, &height);
	if (w->basic.resize || w->core.width == 0)
		w->core.width = width;
	if (w->basic.resize || w->core.height == 0)
		w->core.height = height;
	if (w->raster.image == NULL ||
	    w->raster.image->width != width ||
	    w->raster.image->height != height)
		w->raster.done = False; /* Image size changed, need a new */

	XSetBackground(XtDisplay(w),w->raster.mygc,w->core.background_pixel);
	XSetForeground(XtDisplay(w),w->raster.mygc,w->basic.foreground_pixel);
    }


static void Initialize (request, new)
Widget request, new;
    {
	XeRasterWidget w = (XeRasterWidget)new;

	w->raster.mygc  =
		XCreateGC(XtDisplay(w), DefaultRootWindow(XtDisplay(w)), 0, 0);
	XSetLineAttributes(XtDisplay(w), w->raster.mygc,
			   1, LineSolid, CapButt, JoinMiter);
	w->raster.image = NULL;
	w->raster.num_colors = 0;
	w->raster.num_allocated = 0;
	w->raster.alloc = NULL;
	w->raster.cmap = 0;
	w->raster.data = NULL;
	w->raster.copy = False;
	w->raster.class = XeRasterClass_UNKNOWN;
	w->raster.width = 0;
	w->raster.height = 0;
	w->raster.samples_per_pixel = 0;
	w->raster.bits_per_sample = 0;
	w->raster.color_map = NULL;
	w->raster.resolution_unit = XeResolutionUnit_NONE;
	w->raster.x_resolution = 1.0;
	w->raster.y_resolution = 1.0;
	w->raster.bytes_per_line = 0;
	w->raster.done = False;
	SetupContent(w);
    }

static void Redisplay(w, e, r)
Widget w;
XExposeEvent *e;
Region r;
    {
	XeRasterWidget rw = (XeRasterWidget)w;

	if (!XtIsRealized(w))
		return;
	if (rw->raster.done == False)
		XeRasterLayout(rw);
	if (rw->raster.image != NULL)
	    {
		int srcx = e->x - rw->raster.offset_x;
		int srcy = e->y - rw->raster.offset_y;
		int dstx = e->x;
		int dsty = e->y;
		unsigned int width = e->width;
		unsigned int height = e->height;
		if (srcx < 0)
		    {
			width += srcx;
			srcx = 0;
			dstx = rw->raster.offset_x;
			
		    }
		if (srcy < 0)
		    {
			height += srcy;
			srcy = 0;
			dsty = rw->raster.offset_y;
		    }
		XPutImage(XtDisplay(w), XtWindow(w),
			  rw->raster.mygc, rw->raster.image,
			  srcx, srcy, dstx, dsty,
			  width, height);
	    }
    }


static Boolean SetValues(current, request, new, args, num_args)
Widget current, request, new;
ArgList args;
Cardinal *num_args;
    {
	XeRasterWidget cw = (XeRasterWidget)current;
	XeRasterWidget nw = (XeRasterWidget)new;

	if (cw->raster.visual != nw->raster.visual)
	    {
		nw->raster.visual = cw->raster.visual;
		XeWidgetWarningMsg((Widget)cw, "rasterVisual",
				   "Cannot change visual with set_values",
				   (String *)NULL, 0);
		nw->raster.visual = cw->raster.visual;
	    }
	if (cw->raster.max_colors != nw->raster.max_colors ||
	    cw->raster.colormap_use != nw->raster.colormap_use ||
	    cw->raster.color_quantize != nw->raster.color_quantize)
		nw->raster.done = False;
	SetupContent(nw);
	return (!nw->raster.done ||
		cw->core.background_pixel != nw->core.background_pixel ||
		cw->basic.foreground_pixel != nw->basic.foreground_pixel);
    }

static void Resize(w)
Widget w;
    {
	SetupContent((XeRasterWidget)w);
    }

/*
** Raster widget needs to have own Realize function so that different
** visuals can be specified for it.
*/
static void Realize(widget, value_mask, attributes)
Widget widget;
XtValueMask *value_mask;
XSetWindowAttributes *attributes;
    {
	XeRasterWidget w = (XeRasterWidget)widget;
	Visual *visual;

	if (w->raster.visual == NULL)
		visual = (Visual *)CopyFromParent;
	else
	    {
		visual = w->raster.visual;
		if (visual != DefaultVisualOfScreen(XtScreen(w)))
		    {
			w->raster.cmap = w->core.colormap;
			w->core.colormap = XCreateColormap
				(XtDisplay(w),
				 RootWindowOfScreen(XtScreen(w)),
				 visual,
				 AllocNone);
			attributes->colormap = w->core.colormap;
			*value_mask |= CWColormap;
		    }
	    }
	XtCreateWindow(widget, (unsigned int) InputOutput,
		       visual, *value_mask, attributes);
    }

static void Destroy(w)
Widget w;
    {
	register XeRasterWidget rw = (XeRasterWidget)w;
	
	XFreeGC(XtDisplay(w), rw->raster.mygc);

	if (rw->raster.data != NULL && rw->raster.copy)
		free((void *)rw->raster.data);
	if (rw->raster.color_map != NULL)
		XtFree((char *)rw->raster.color_map);
	XeRasterFreeColors(rw);
	if (rw->raster.image != NULL)
	    {
		XtFree(rw->raster.image->data);
		XFree(rw->raster.image);
	    }
    }

/* Should probably make Notify in Basic global and call it.. */

static void Notify(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
    {
	XeNotifyCallbackData call_data;
	call_data.reason = XeCR_NOTIFY;
	call_data.event = event;
	call_data.params = params;
	call_data.num_params = *num_params;
	XtCallCallbackList(w, ((XeBasicWidget)w)->basic.notify_callbacks,
			   (XtPointer)&call_data);
    }

/*
** Widget needs to follow cursor to change the colourmaps if private
** colourmaps are installed.
**
** A private colormap is indicated by non-zero value in raster.cmap
** field (which is actually the saved old default colormap).
*/
static void EnterWindow(w, e, s, n)
Widget w;
XEvent *e;
String *s;
Cardinal *n;
    {

	XeRasterWidget r = (XeRasterWidget)w;
	Widget wlst[2];
	
	if (r->raster.cmap == 0)
		return;	/* No private colormap, nothing to do */
	if (r->raster.colormap_install)
	    {
		XInstallColormap(XtDisplay(w), r->core.colormap);
		return;
	    }
	while (!XtIsShell(w))
		w = XtParent(w);
	wlst[0] = (Widget)r;
	wlst[1] = w;
	XtSetWMColormapWindows(w, wlst, (Cardinal)2);
	/* XSetWindowColormap(XtDisplay(w), XtWindow(w), r->core.colormap); */
    }

static void LeaveWindow(w, e, s, n)
Widget w;
XEvent *e;
String *s;
Cardinal *n;
    {
	XeRasterWidget r = (XeRasterWidget)w;
	if (r->raster.cmap == 0)
		return;	/* No private colormap, nothing to do */
	if (r->raster.colormap_install)
	    {
		XUninstallColormap(XtDisplay(w), r->core.colormap);
		return;
	    }
	while (!XtIsShell(w))
		w = XtParent(w);
	XtSetWMColormapWindows(w, &w, (Cardinal)1);
	/* XSetWindowColormap(XtDisplay(w), XtWindow(w), w->core.colormap); */
    }
