/*
** RasterImport module for importing raster images into
** Xew internal format. This module is a generic switching
** point to different format specific import functions and
** is a place for common actions necessary for import.
**
** Copyright 1992 by Markku Savela and
**	Technical Research Centre of Finland
**
** The author maybe contacted via:
**
**    Mail:	Markku Savela
**		Technical Research Centre of Finland
**		Telecommunications Laboratory
**		Otakaari 7 B, P.O. Box 34
**		SF-02151 ESPOO
**		Finland
**
**    Phone:	+358 0 456 6509
**    Email:	savela@tel.vtt.fi
*/
#if SYSV_INCLUDES
#	include <malloc.h>	/* malloc() et al. */
#endif
#if ANSI_INCLUDES
#	include <stdlib.h>
#endif
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xew/RasterP.h>
#include "RasterImport.h"
#include "RasterTIFF.h"
#include "RasterGIF.h"
#include "RasterJPEG.h"

static void RasterImportWarning(w, msg)
XeRasterWidget w;
char *msg;
    {
	XeWidgetWarningMsg
		((Widget)w, "rasterImportWarning", msg, (String *)NULL, 0);
    }

      
void XeRasterImport(w)
XeRasterWidget w;
    {
	if (w->raster.data != NULL && w->raster.copy)
		free((void *)w->raster.data);
	w->raster.data = NULL;
	if (w->raster.color_map != NULL)
		XtFree((char *)w->raster.color_map);
	w->raster.color_map = NULL;
	switch (w->basic.content_format)
	    {
	    case XeContentFormat_TIFF:
		XeImport_TIFF(w);
		break;
	    case XeContentFormat_GIF:
		XeImport_GIF(w);
		break;
	    case XeContentFormat_JPEG:
		XeImport_JPEG(w);
		break;
	    default:
		XeWidgetWarningMsg
			((Widget)w, "formatNotSupported",
			 "Raster widget does not support this format",
			 (String *)NULL, 0);
		break;
	    }
    }

