#ifndef _XewRasterP_h
#define _XewRasterP_h
/*
** Copyright 1992, 1993 by Markku Savela and
**	Technical Research Centre of Finland
**
** The author may be contacted via:
**
**    Mail:	Markku Savela
**		Technical Research Centre of Finland
**		Telecommunications Laboratory
**		Otakaari 7 B, P.O. Box 34
**		SF-02151 ESPOO
**		Finland
**
**    Phone:	+358 0 456 6509
**    Email:	savela@tel.vtt.fi
*/
#include <X11/Xew/BasicP.h>
#include <X11/Xew/Raster.h>

typedef struct _XeRasterClass
    {
	int makes_compiler_happy;	/* not used */
    } XeRasterClassPart;

typedef struct _XeRasterClassRec
    {
	CoreClassPart	core_class;
	XeBasicClassPart basic_class;
	XeRasterClassPart raster_class;
    } XeRasterClassRec;

typedef enum
    {
	XeRasterClass_UNKNOWN,	/* Unspecified type, error */
	XeRasterClass_BILEVEL,	/* bi-level (black and white) image */
	XeRasterClass_GRAYSCALE,/* Grayscale image */
	XeRasterClass_PALETTE,	/* Palette colour image */
	XeRasterClass_FULLCOLOR	/* Full color image */
    } XeRasterClass;

typedef enum
    {
	XeResolutionUnit_NONE,	/* Bare numbers */
	XeResolutionUnit_INCH,	/* Inch */
	XeResolutionUnit_CM    	/* Centimeters */
    } XeResolutionUnit;

typedef struct
    {
	/* resources */

	unsigned int clip_x, clip_y;
	unsigned int clip_width, clip_height;
	Dimension offset_x, offset_y;
	int max_colors;		/* Requested max number of different colours */
	XeColormapUse colormap_use;
	XeColorQuantize color_quantize;
	/*
	** colormapInstall resource is only needed to support Window
	** managers that don't work with XtSetWMColormapWindows. If
	** True, the widget will explicitly install colormaps
	** on enter/leave events.
	*/
	Boolean colormap_install;

	/* private state */

	Visual *visual;		/* The Visual associated with the window */

	/* The Current display version of the Image */

	GC mygc;		/* Graphic Context */
	XImage *image;		/* Display Image */
	int num_colors;		/* Actual count of different colors in image */
	int num_allocated;	/* Number of allocated colors */
	unsigned long *alloc;	/* List of allocated colors */
	Colormap cmap;		/* Private Colormap */

	/* The full data of the current Image. These fields *must* be
	** filled in by the RasterImport module and they are independent
	** on the capabilities of the current display. The import
	** function is responsible for converting the raster image
	** into the format described below.
	*/
	unsigned char *data;	/* Pointer to raw bitmap data */
	unsigned int copy:1;	/* TRUE, if data malloc'ed */
	unsigned int done:1;	/* TRUE, when XeRasterLayout done */
	XeRasterClass class;	/* Raster Image class */
	XeResolutionUnit resolution_unit;
	float x_resolution, y_resolution;
	unsigned int
		width,		/* Number of columns in raster image */
		height;		/* Number of rows in raster image */
	/*
	** The following parameters describe the structure of the
	** raw image data:
	**
	** samples_per_pixel	Number of components in pixel value.
	** bits_per_sample	Number of significant bits in each
	**			component sample.
	** bits_per_component	Number of actual bits allocated for
	**			each component (e.g. samples may be
	**			only in range 0-15, but each component
	**			might be allocated one byte).
	** bytes_per_line	Total number of bytes in single pixel
	**			row (this includes all necessary padding).
	*/
	unsigned short
		samples_per_pixel, bits_per_sample,
		bits_per_component;
	int bytes_per_line;
	unsigned char *color_map;/* Size =  3 * (1<<bits_per_sample) bytes */
    } XeRasterPart;

typedef struct _XeRasterRec
    {
	CorePart	core;
	XeBasicPart	basic;
	XeRasterPart	raster;
    } XeRasterRec;

/* Class pointer */
extern XeRasterClassRec xeRasterClassRec;

#endif




