/*
** A demo program for TextEd widget
**
** WARNING!
**	This is not any real "editor". The user interface is just a
**	jumbled collection of buttons and menus which access the
**	TextEd features individually at low level.
*/
#include <stdio.h>
#include <string.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xew/Basic.h>
#include <X11/Xew/Raster.h>
#include <X11/Xew/Audio.h>
#include <X11/Xew/TextEd.h>
#include <X11/Xew/Support.h>
#if SYSV_INCLUDES || ANSI_INCLUDES
#	include <limits.h>
#else
#ifndef LONG_MAX
#define LONG_MAX (~((unsigned long)0L)>>1)
#endif
#endif

static Widget root_widget;
static Widget edit_widget;	/* TextEd widget instance */
static char *edit_file;		/* Current primary file */
static Widget edit_title;	/* Label Widget for filename */

typedef enum
    {
	FileMenu_None,
	FileMenu_SaveAs,
	FileMenu_LoadFile
    } FileMenuAction;


static Position AdjustCoord(x, y, a, b)
int x, a, b;
    {
	if (x < a)
		a = x;
	else if (x + y >= a + b)
		a = x + y - b;
	return a;
    }

/*
** Sample Notify skeleton..
*/
static void Notify(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	XeNotifyCallbackData *notify = (XeNotifyCallbackData *)call_data;
	XeNotifyAreaCallbackData *area;
	int i;
	Position x, y;

	switch (notify->reason)
	    {
	    case XeCR_NOTIFY:
		/* Quickhack: just assume it's a button event! */
		printf("Notify(");
		for (i = 0; i < notify->num_params; i++)
			printf("%s%s", i==0?"":",", notify->params[i]);
		printf(") at (%d,%d)\n",
		       (int)notify->event->xbutton.x,
		       (int)notify->event->xbutton.y);
		break;
	    case XeCR_NOTIFY_AREA:
		area = (XeNotifyAreaCallbackData *)call_data;
		x = AdjustCoord(area->area.x, area->area.width,
				area->visible.x, area->visible.width);
		y = AdjustCoord(area->area.y, area->area.height,
				area->visible.y, area->visible.height);
		if (x != area->visible.x || y != area->visible.y)
			XawViewportSetCoordinates(XtParent(w), x, y);
		break;
	    default:
		printf("UNKNOWN: XeRC=%d\n");
		break;
	    }
    }

static void Quit(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	XtDestroyApplicationContext(XtWidgetToApplicationContext(w));
	exit(0);
    }

static void SelectAll(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	XeTextSetSelection(edit_widget, (long)0, (long)LONG_MAX);
    }

/*
** Generic PopDown function
*/
static void PopDown(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	while (!XtIsShell(w))
		w = XtParent(w);
	XtPopdown(w);
    }

/*
** SetFilenameToTitle
*/
static void SetFilenameToTitle()
    {
	Arg args[1];
	char *label = edit_file ? edit_file : "(noname)";

	XtSetArg(args[0], XtNlabel, label);
	XtSetValues(edit_title, args, (Cardinal)1);
    }

/*
** File dialog is copied and modified from X11R5 editres client code!
*/
static void  PopdownFileDialog(w, client_data, junk)
Widget w;
XtPointer client_data, junk;
    {
	Widget dialog = XtParent(w);
	
	if ((FileMenuAction)client_data)
	    {
		String filename = XawDialogGetValueString(dialog);
		Cardinal n;
		Arg args[2];

		switch ((FileMenuAction)client_data)
		    {
		    case FileMenu_SaveAs:
			XeTextSaveAsFile(edit_widget, filename);
			break;
		    case FileMenu_LoadFile:
			n = 0;
			XtSetArg(args[n], XtNcontentFile, filename); n++;
			XtSetArg(args[n], XtNcontentLoaded, False); n++;
			XtSetValues(edit_widget, args, n);
			break;
		    default:
			break;
		    }
		if (edit_file)
			free(edit_file);
		edit_file = strdup(filename);
		SetFilenameToTitle();
	    }
	XtPopdown(XtParent(dialog));
    }

static void SaveToFile(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	XeTextSaveAsFile(edit_widget, edit_file);
    }


static void PopupCentered(event, w, mode)
XEvent * event;
Widget w;
XtGrabKind mode;
    {
	Boolean get_from_cursor = FALSE;
	Arg args[3];
	Cardinal num_args;
	Dimension width, height, b_width;
	int x, y, max_x, max_y;
	
	XtRealizeWidget(w);
	
	if (event == NULL)
		get_from_cursor = TRUE;
	else
	    {
		switch (event->type)
		    {
		    case ButtonPress:
		    case ButtonRelease:
			x = event->xbutton.x_root;
			y = event->xbutton.y_root;
			break;
		    case KeyPress:
		    case KeyRelease:
			x = event->xkey.x_root;
			y = event->xkey.y_root;
			break;
		    default:
			get_from_cursor = TRUE;
			break;
		    }
	    }
	if (get_from_cursor)
	    {
		Window root, child;
		int win_x, win_y;
		unsigned int mask;
		
		XQueryPointer(XtDisplay(w), XtWindow(w),
			      &root, &child, &x, &y, &win_x, &win_y, &mask);
	    }
	num_args = 0;
	XtSetArg(args[num_args], XtNwidth, &width); num_args++;
	XtSetArg(args[num_args], XtNheight, &height); num_args++;
	XtSetArg(args[num_args], XtNborderWidth, &b_width); num_args++;
	XtGetValues(w, args, num_args);
	width += 2 * b_width;
	height += 2 * b_width;
	x -= ((int) width/2);
	if (x < 0) 
		x = 0;
	if ( x > (max_x = (int) (XtScreen(w)->width - width)) )
		x = max_x;
	y -= ( (Position) height/2 );
	if (y < 0) 
		y = 0;
	if ( y > (max_y = (int) (XtScreen(w)->height - height)) )
		y = max_y;
	num_args = 0;
	XtSetArg(args[num_args], XtNx, x); num_args++;
	XtSetArg(args[num_args], XtNy, y); num_args++;
	XtSetValues(w, args, num_args);
	XtPopup(w, mode);
    }

typedef struct DialogDef
    {
	char *label;
	char **value;
	XtCallbackProc callback;
	XtPointer okay_data;
	Widget shell;
	Widget dialog;
    } DialogDef;

static DialogDef popup_dialog_saveas =
    {
	"File name: ",
	&edit_file,
	PopdownFileDialog,
	(XtPointer)FileMenu_SaveAs,
    };

static DialogDef popup_dialog_loadfile =
    {
	"File name: ",
	&edit_file,
	PopdownFileDialog,
	(XtPointer)FileMenu_LoadFile,
    };

static void PopupDialog(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	Cardinal n;
	Arg args[2];
	DialogDef *def = (DialogDef *)client_data;
	
	if (!def->shell)
	    {
		def->shell = XtCreatePopupShell
			("dialog", transientShellWidgetClass,
			 XtParent(w), NULL, 0);
		n = 0;
		if (def->label)
			XtSetArg(args[n], XtNlabel, def->label); n++;
		if (def->value)
		    {
			if (*def->value)
				XtSetArg(args[n], XtNvalue, *def->value);
			else
				XtSetArg(args[n], XtNvalue, "");
			n++;
		    }
		def->dialog=XtCreateManagedWidget("dialog", dialogWidgetClass, 
						  def->shell, args, n);
		if (def->okay_data)
		    {
			XawDialogAddButton(def->dialog,"okay",def->callback,
					   def->okay_data);
			XawDialogAddButton(def->dialog,"cancel",def->callback,
					   (XtPointer)0);
		    }
		else
			XawDialogAddButton(def->dialog,"OK", def->callback,
					   (XtPointer)0);
			
	    }
	PopupCentered(NULL, def->shell, XtGrabNone);
    }

static void WarningMsgHandler(name, type, class, defaultp, params, num_params)
String name;
String type;
String class;
String defaultp;
String *params;
Cardinal *num_params;
    {
	static char message[1000];

	static DialogDef popup_errormsg =
	    {
		message,
		NULL,
		PopDown,
		(XtPointer)NULL,
	    };


	char buffer[1000];
	XtGetErrorDatabaseText(name,type,class,defaultp,buffer,sizeof(buffer));
	if (params && num_params && *num_params > 0)
	    {
		int i = *num_params, j;
		String par[10];
		if (i > 10)
			i = 10;
		for (j = 0; j < i; j++)
			par[j] = params[j];
		while (i < 10)
			par[i++] = NULL;
		sprintf(message, buffer, 
			par[0], par[1], par[2], par[3], par[4],
			par[5], par[6], par[7], par[8], par[9]);
	    }
	else
		strcpy(message, buffer);
	if (popup_errormsg.dialog)
	    {
		Arg args[1];
		XtSetArg(args[0], XtNlabel, message);
		XtSetValues(popup_errormsg.dialog, args, (Cardinal)1);
	    }
	PopupDialog(edit_title, (XtPointer)&popup_errormsg,
		    (XtPointer)NULL);
    }

static Widget CreateToggleElement(parent, name, args, n)
Widget parent;
char *name;
Arg *args;
int n;
    {
	Widget box, toggle, label;

	XtSetArg(args[n], XtNorientation, XtorientHorizontal); ++n;
	XtSetArg(args[n], XtNborderWidth, 0); ++n;
	XtSetArg(args[n], XtNhSpace, 1); ++n;
	XtSetArg(args[n], XtNvSpace, 1); ++n;
	XtSetArg(args[n], XtNborderWidth, 0); ++n;
	box = XtCreateManagedWidget
		("toggleBox", boxWidgetClass, parent, args, n);
	n = 0;
	XtSetArg(args[n], XtNlabel, " "); ++n;
	toggle = XtCreateManagedWidget
		("toggleButton", toggleWidgetClass, box, args, n);
	n = 0;
	XtSetArg(args[n], XtNborderWidth, 0); ++n;
	label = XtCreateManagedWidget(name, labelWidgetClass, box, args, n);
	return toggle;
    }


typedef enum
    {
	TOGGLE,	/* Plain toggle element */
	RADIO,	/* Toggle element, add to radio-group of previous toggle */
	PANE	/* A new pane */
    } ToggleControl;

typedef struct ToggleDef
    {
	char *name;
	ToggleControl control;
	char *data;
	Widget toggle;
    } ToggleDef;

ToggleDef format_control[] =
    {
	{" ",			PANE,	XtNformat,},
	{"wrap lines",		TOGGLE,	(char *)True,},
	{"don't wrap lines",	RADIO,	(char *)False,},
	{"Alignment:",		PANE,	XtNalignment,},
	{"start (ragged right)",TOGGLE,	(char *)XeAlignment_START,},
	{"end (ragged left)",	RADIO,	(char *)XeAlignment_END,},
	{"centered",		RADIO,	(char *)XeAlignment_CENTER,},
	{"justified",		RADIO,	(char *)XeAlignment_JUSTIFIED,},
	{"Export format:",	PANE,	XtNexportFormat,},
	{"string",		TOGGLE,	(char *)XeTextExport_STRING,},
	{"string (F)",		RADIO,	(char *)XeTextExport_STRING_F,},
	{"character odif",	RADIO,	(char *)XeTextExport_ODIF,},
	{"character odif (F)",	RADIO,	(char *)XeTextExport_ODIF_F,},
	{"character odif (FP)",	RADIO,	(char *)XeTextExport_ODIF_FP,},
	{NULL},
    };

static void SetResourcesFromToggles(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	ToggleDef *r = (ToggleDef *)client_data;
	char *resource_name = NULL;

	for (; r->name; ++r)
		switch (r->control)
		    {
		    case PANE:
			resource_name = r->data;
			break;
		    case TOGGLE:
		    case RADIO:
			if (r->toggle && resource_name)
			    {
				Arg args[1];
				Boolean state;

				XtSetArg(args[0], XtNstate, &state);
				XtGetValues(r->toggle, args, 1);
				if (state)
				    {
					XtSetArg(args[0],resource_name,r->data);
					XtSetValues(edit_widget, args, 1);
				    }
			    }
			break;
		    default:
			break;
		    }
    }


ToggleDef insert_state[] =
    {
	/* *NOTE* The control sequences in data assume the C compiler is
	   is using ASCII code set. --msa */
	{" ",			PANE,	NULL,},
	{"Normal intensity",	TOGGLE,	"\033[22m",},
	{"Bold",		RADIO,	"\033[1m",},
	{"Italic",		TOGGLE,	"\033[3m",},
	{"Not italic",		RADIO,	"\033[23m",},
	{"Underlined", 		TOGGLE,	"\033[4m",},
	{"Double underlined",	RADIO,	"\033[21m",},
	{"Not underlined",	RADIO,	"\033[24m",},
	{"Inverse", 		TOGGLE,	"\033[7m",},
	{"Not inversed", 	RADIO,	"\033[27m",},
	{"Crossed out", 	TOGGLE,	"\033[9m",},
	{"Not crossed out", 	RADIO,	"\033[29m",},
	{"Superscript",		TOGGLE,	"\033L\033L"},
	{"Baseline",		RADIO,	"\033K\033K\033L"},
	{"Subscript",		RADIO,	"\033K\033K"},
	{"Foreground:",		PANE,	NULL,},
	{"Default",		TOGGLE,	"\033[39m",},
	{"Black",		RADIO,	"\033[30m",},
	{"Red",			RADIO,	"\033[31m",},
	{"Green",		RADIO,	"\033[32m",},
	{"Yellow",		RADIO,	"\033[33m",},
	{"Blue",		RADIO,	"\033[34m",},
	{"Magenta",		RADIO,	"\033[35m",},
	{"Cyan",		RADIO,	"\033[36m"},
	{"White",		RADIO,	"\033[37m"},
	{"Background:",		PANE,	NULL,},
	{"Default",		TOGGLE,	"\033[49m",},
	{"Black",		RADIO,	"\033[40m",},
	{"Red",			RADIO,	"\033[41m",},
	{"Green",		RADIO,	"\033[42m",},
	{"Yellow",		RADIO,	"\033[43m",},
	{"Blue",		RADIO,	"\033[44m",},
	{"Magenta",		RADIO,	"\033[45m",},
	{"Cyan",		RADIO,	"\033[46m"},
	{"White",		RADIO,	"\033[47m"},
	{"Font:",		PANE,	NULL,},
	{"primary (default)",	TOGGLE,	"\033[10m"},
	{"First alternative",	RADIO,	"\033[11m"},
	{"Second alternative",	RADIO,	"\033[12m"},
	{"Third alternative",	RADIO,	"\033[13m"},
	{"Fourth alternative",	RADIO,	"\033[14m"},
	{"Fifth alternative",	RADIO,	"\033[15m"},
	{"Sixth alternative",	RADIO,	"\033[16m"},
	{"Seventh alternative",	RADIO,	"\033[17m"},
	{"Eight alternative",	RADIO,	"\033[18m"},
	{"Ninth alternative",	RADIO,	"\033[19m"},
	{"Character set:",	PANE,	NULL,},
	{"iso8859-1 (Latin No.1)",TOGGLE,"\33(B\33)B\33.A\33/A\17\33\155"},
	{"iso8859-2 (Latin No.2)",RADIO,"\33(B\33)B\33.B\33/B\17\33\155"},
	{"iso8859-3 (Latin No.3)",RADIO,"\33(B\33)B\33.C\33/C\17\33\155"},
	{"iso8859-4 (Latin No.4)",RADIO,"\33(B\33)B\33.D\33/D\17\33\155"},
	{"iso8859-5 (Latin/Cyrillic)",RADIO,"\33(B\33)B\33.L\33/L\17\33\155"},
	{"iso8859-6 (Latin/Arabic)",RADIO,"\33(B\33)B\33.G\33/G\17\33\155"},
	{"iso8859-7 (Latin/Greek)",RADIO,"\33(B\33)B\33.F\33/F\17\33\155"},
	{"iso8859-8 (Latin/Hebrew)",RADIO,"\33(B\33)B\33.H\33/H\17\33\155"},
	{"iso8859-9 (Latin No.5)",RADIO,"\33(B\33)B\33.M\33/M\17\33\155"},
	{"iso6937",		RADIO,	"\33(B\33)B\33.R\33/R\17\33\155"},
	{"iso646-se",		RADIO,	"\33(G\33)G\33.A\33/A\17\33\155"},
	{"iso646-se (names)",	RADIO,	"\33(H\33)H\33.A\33/A\17\33\155"},
	{"JIS C 6226-1983",	RADIO,	"\33(B\33)B\33$*B\33$+B\17\33\155"},
	{"KSC 5601-1987",	RADIO,	"\33(B\33)B\33$*C\33$+C\17\33\155"},
	{"GB 2312-80",		RADIO,	"\33(B\33)B\33$*A\33$+A\17\33\155"},
	{NULL},
    };

static char *GetCurrentInsertState()
    {
	static char *current;
	static int size;

	struct ToggleDef *r;
	Arg args[1];
	Boolean state;
	char *s;
	int i;

	for (i = 0, r = insert_state; r->name; ++r)
		switch (r->control)
		    {
		    case TOGGLE:
		    case RADIO:
			if (!r->toggle || !r->data)
				break;
			XtSetArg(args[0], XtNstate, &state);
			XtGetValues(r->toggle, args, 1);
			if (state)
			    {
				int length = strlen(r->data);

				if (i + length >= size)
				    {
					size = (size + length + 50) / 50 * 50;
					current = XtRealloc(current, size);
				    }
				strcpy(current + i, r->data);
				i += length;
			    }
			break;
		    default:
			break;
		    }
	return i > 0 ? current : NULL;
    }

static void SetInsertState(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	Arg args[1];
	long start, end;

	XtSetArg(args[0], XtNinitialState, GetCurrentInsertState());
	XtSetValues(edit_widget, args, 1);
	/*
	** If a selection is currently active, then extract the selected
	** text away from the widget, and then replace the text with the
	** extracted text. This is rather inefficient way of getting the
	** attributes of the selected text to change as defined by the
	** input state.
	**
	** NOTE:As the selected content is effected by the currently
	**	chosen export format, some information may be lost,
	**	if the format is too simple (like string).
	*/
	XeTextGetSelectionPosition(edit_widget, &start, &end);
	if (end > start)
	    {
		long length;
		String value;

		value = XeTextGetSubstring(edit_widget, &length, start, end);
		XeTextReplace(edit_widget, start, end, value, length);
		XeTextUnsetSelection(edit_widget);
		XtFree(value);
	    }
    }

static Widget CreateTogglePanel(parent, name, definition)
Widget parent;
char *name;
ToggleDef *definition;
    {
	Arg args[10];
	int n;
	Widget panel, pane, radio = 0;
	char *s;
	ToggleDef *r;

	n = 0;
	XtSetArg(args[n], XtNinternalBorderWidth, 0); ++n;
	XtSetArg(args[n], XtNorientation, XtorientHorizontal); ++n;
	pane = panel = XtCreateManagedWidget
		(name, panedWidgetClass, parent, args, n);
	for (r = definition; r->name; r++)
		switch (r->control)
		    {
		    case PANE:
			n = 0;
			XtSetArg(args[n], XtNinternalBorderWidth, 0); ++n;
			XtSetArg(args[n], XtNshowGrip, False); ++n;
			pane = XtCreateManagedWidget
				(name, panedWidgetClass, panel, args, n);
			n = 0;
			XtSetArg(args[n], XtNborderWidth, 0); ++n;
			XtSetArg(args[n], XtNshowGrip, False); ++n;
			r->toggle = XtCreateManagedWidget
				(r->name, labelWidgetClass, pane, args, n);
			break;
		    case TOGGLE:
		    case RADIO:
			n = 0;
			XtSetArg(args[n], XtNshowGrip, False); ++n;
			r->toggle = CreateToggleElement(pane,r->name,args,n);
			if (r->data)
			    {
				n = 0;
				XtSetArg(args[n], XtNradioData, r->data); ++n;
				XtSetValues(r->toggle, args, n);
			    }
			if (r->control == RADIO)
				XawToggleChangeRadioGroup(r->toggle, radio);
			radio = r->toggle;
			break;
		    default:
			break;
		    }
	return panel;
    }

static Widget CreateBottomPanel(parent, name, set_callback, set_data)
Widget parent;
char *name;
XtCallbackProc set_callback;
XtPointer set_data;
    {
	int n;
	Arg args[3];
	Widget box, but;

	n = 0;
	XtSetArg(args[n], XtNshowGrip, False); ++n;
	XtSetArg(args[n], XtNborderWidth, 0); ++n;
	box = XtCreateManagedWidget(name, boxWidgetClass,parent,args,n);
	n = 0;
	but = XtCreateManagedWidget("set", commandWidgetClass, box, args, n);
	XtAddCallback(but, XtNcallback, set_callback, set_data);
	n = 0;
	but = XtCreateManagedWidget("close",commandWidgetClass,box,args,n);
	XtAddCallback(but, XtNcallback, PopDown, (XtPointer)parent);
    }

typedef struct PanelDef
    {
	char *name;
	ToggleDef *toggles;
	XtCallbackProc set_callback;
	Widget shell;
    } PanelDef;

static PanelDef popup_format_control =
    {
	"Change Layout and Export",
	format_control,
	SetResourcesFromToggles,
    };

static PanelDef popup_insert_state =
    {
	"Change Input State",
	insert_state,
	SetInsertState,
    };

static void PopupPanel(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
    {
	PanelDef *panel = (PanelDef *)client_data;
	Widget close, pane;
	Cardinal n;
	Arg args[2];

	if (!panel->shell)
	    {
		panel->shell = XtCreatePopupShell
			(panel->name, transientShellWidgetClass,
			 XtParent(w), NULL, 0);
		n = 0;
		pane = XtCreateManagedWidget
			("panel", panedWidgetClass, panel->shell, args, n);
		n = 0;
		XtSetArg(args[n], XtNborderWidth, 0); ++n;
		XtSetArg(args[n], XtNshowGrip, False); ++n;
		(void)XtCreateManagedWidget
			(panel->name,labelWidgetClass, pane, args, n);
		CreateTogglePanel(pane, "toggles", panel->toggles);
		CreateBottomPanel(pane, "bottom", panel->set_callback,
				  (XtPointer)panel->toggles);
	    }
	PopupCentered(NULL, panel->shell, XtGrabNone);
    }

typedef struct MenuDef
    {
	char *name;
	XtCallbackProc callback;
	XtPointer client_data;
    } MenuDef;

static MenuDef file_menu[] =
    {
	{"Save", SaveToFile, (XtPointer)NULL,},
	{"Save as", PopupDialog, (XtPointer)&popup_dialog_saveas,},
	{"Load File", PopupDialog, (XtPointer)&popup_dialog_loadfile,},
	{"Quit", Quit, NULL},
	{NULL},
    };


static MenuDef edit_menu[] =
    {
	{"Character format", PopupPanel, (XtPointer)&popup_insert_state,},
	{"Layout format", PopupPanel, (XtPointer)&popup_format_control,},
	{"Select all", SelectAll, (XtPointer)NULL},
	{NULL},
    };


static void CreateMenu(name, parent, definition)
char *name;
Widget parent;
MenuDef *definition;
    {
	Widget menu, entry, button;

	button = XtCreateManagedWidget
		(name, menuButtonWidgetClass, parent, NULL, 0);
	menu = XtCreatePopupShell
		("menu", simpleMenuWidgetClass, button, NULL, 0);
	for ( ;definition->name; definition++)
	    {
		entry = XtCreateManagedWidget
			(definition->name, smeBSBObjectClass, menu, NULL, 0);
		XtAddCallback(entry, XtNcallback, definition->callback,
			      definition->client_data);
	    }
    }

main(argc, argv)
int argc; char *argv[];
    {
	XtAppContext demo_application;
	Widget view, widget, pane, box, button;
	Arg args[10];
	int n;
	XeContentFormat content_type;

	root_widget = XtAppInitialize(&demo_application, "TextEd",
				      NULL,0, &argc, argv, (String *)NULL,
				      NULL, 0);

	edit_file = argc > 1 ? strdup(argv[1]) : NULL;

	n = 0;
	pane = XtCreateManagedWidget
		("pane", panedWidgetClass, root_widget, args, n);
	n = 0;
	box = XtCreateManagedWidget
		("box", boxWidgetClass, pane, args, n);

	CreateMenu("File", box, file_menu);
	CreateMenu("Edit", box, edit_menu);

	n = 0;
	XtSetArg(args[n], XtNborderWidth, 0); ++n;
	edit_title = XtCreateManagedWidget("fileName", labelWidgetClass,
					  box, args, n);
	SetFilenameToTitle();
	XtAppSetWarningMsgHandler(demo_application, WarningMsgHandler);
	n = 0;
	XtSetArg(args[n], XtNallowHoriz, True); ++n;
	XtSetArg(args[n], XtNallowVert, True); ++n;
	XtSetArg(args[n], XtNshowGrip, False); ++n;
	view = XtCreateManagedWidget("demoTextEd", viewportWidgetClass,
				     pane,args, n);
	content_type = XeFindContentFormat(edit_file);
	n = 0;
	XtSetArg(args[n], XtNcontentFile, edit_file); ++n;
	XtSetArg(args[n], XtNcontentFormat, content_type); ++n;
	if (content_type == XeContentFormat_UNKNOWN)
	    {
		/*
		** Request proportional spacing, because we are not
		** specifying the lineSpacing.
		*/
		XtSetArg(args[n], XtNproportional, True); ++n;
		XtSetArg(args[n], XtNcolumnWidth, 400); ++n;
		edit_widget = XtCreateManagedWidget
			("TextEdit", xeTextEdWidgetClass, view, args, n);
	    }
	else if (content_type == XeContentFormat_AUDIO)
	    {
		XtSetArg(args[n], XtNplay, XePlay_FREEZE); ++n;
		edit_widget = XtCreateManagedWidget
			("AudioEdit", xeAudioWidgetClass, view, args, n);
	    }
	else
	    {
		XtSetArg(args[n], XtNscaling, XeScaling_NONE); ++n;
		edit_widget = XtCreateManagedWidget
			("EditRaster", xeRasterWidgetClass, view, args, n);
	    }
	XtAddCallback(edit_widget, XtNnotifyCallback, Notify, (XtPointer)NULL);
	XtRealizeWidget(root_widget);
	XtAppMainLoop(demo_application);
	return 0;
    }
